import os
import numbers

from . import file_ops as fo
from . import imagestacker as ims


def parse_arguments():
    import argparse
    parser = argparse.ArgumentParser()
    # TODO: excess images in directory? Insufficient images in directory?
    parser.add_argument("--dirs_in", dest="dirs_in", nargs="*", type=str, help="List any directories you want to use.")
    parser.add_argument("--files_in", dest="files_in", nargs="*", type=str,
                        help="List any images or videos you want to use. Not compatible"
                             "with -to_imgs, dirs_in. If -to_vid, must be videos.")
    parser.add_argument("--ext_in", dest="ext_in", type=str, choices=["png", "jpg", "mp4"],
                        help="Will select files with given extension for input.")
    parser.add_argument("--ext_out", dest="ext_out", type=str, choices=["png", "jpg", "mp4"],
                        help="Outputs file with given extension."
                             " Overriden by -to_vid, -to_img, and -to_imgs")
    parser.add_argument("--dir_out", dest="dir_out", type=str, default="./output/", help="Output directory.")
    parser.add_argument("--name", dest="name", type=str, default='file', help="File name. Do not include in dir_out")
    parser.add_argument("--to_vid", dest="to_vid", action='store_true',
                        help="Will output a video file (default 30fps, .mp4). Not compatible with --to_img."
                             " If multiple directories are provided, will only use first x*y videos."
                             " If one directory is provided, will use first x*y videos.")
    parser.add_argument("--to_img", dest="to_img", action='store_true',
                        help="Will output an image file (default .jpg). Not compatible with --to_vid")
    parser.add_argument("--to_imgs", dest="to_imgs", action='store_true',
                        help="Will output many image files (default .jpg)")

    parser.add_argument("--resize_in", dest="resize_in", nargs=2, type=int,
                        help="Sets the dimensions of the input image. "
                             "Not compatible with -resize_out or -resize_down or -resize_up")
    parser.add_argument("--resize_out", dest="resize_out", nargs=2, type=int,
                        help="Sets the dimensions of the output image. "
                             "Not compatible with -resize_in or -resize_down or -resize_up")
    parser.add_argument("--resize_up", dest="resize_up", action='store_true',
                        help="Resizes all input images to the largest image in the set. "
                             "Computed by area of image (eg. width * height). Will override --resize_in, --resize_out."
                             " Not compatible with -resize_down, -resize_first.")
    parser.add_argument("--resize_down", dest="resize_down", action='store_true',
                        help="Resizes all input images to the smallest image in the set. "
                             "Computed by area of image (eg. width * height). Will override --resize_in, --resize_out."
                             " Not compatible with --resize_up, --resize_first")
    parser.add_argument("--resize_first", dest="resize_first", action='store_true',
                        help="Resizes all input images to first small image in the set. --resize_in, --resize_out. "
                             " Not compatible with --resize_up, --resize_down.")

    parser.add_argument("--cols", dest="cols", type=int, default=1,
                        help="Number of images or videos placed side by side, horizontally.")
    parser.add_argument("--rows", dest="rows", type=int, default=1,
                        help="Number of images or videos stacked on top of each other,"
                             " vertically.")

    parser.add_argument("--fps", dest="fps", default=30, type=int,
                        help="Frame rate of video. Not compatible if videos are passed in.")
    parser.add_argument("--max", dest="max_imgs", default=None, type=int,
                        help="Maximum number of images to output (eg. if folder has 1000 images, output only 10.")
    parser.add_argument("--read_matching_file_names", action="store_true",
                        help="Will concatenate files with the same name from each directory,"
                             " and will resize on a per image basis unless a width and height are specified.")

    args = parser.parse_args()
    args.dir_out = fo.append_forward_slash_path(args.dir_out)
    args.dirs_in = fo.append_forward_slash_path(args.dirs_in)
    validate_arguments(args)
    if args.ext_in == "mp4":
        args.to_vid = True
    # Set default extension if none given
    if args.ext_in is None and args.dirs_in is not None:
        args.ext_in = ["png", "jpg"]
        print("Input extension set to %s." % (",".join(args.ext_in),))

    if args.to_vid:
        if args.ext_out not in ['mp4']:
            args.ext_out = 'mp4'
            print("Output extension automatically set to %s." % args.ext_in)

    if args.ext_out is None:
        args.ext_out = "jpg"
        print("Output extension automatically set to %s." % args.ext_out)

    print(args)
    return args


def get_ext(args):
    if args.files_in is not None:
        ext_in = os.path.splitext(args.files_in[0])[1]
        return ext_in[1:]


# Checks if the extensions are equivalent.
def check_ext(ext1: str, ext2: str):
    if ext1.lower() == ext2.lower():
        return True
    elif ext1.lower() in [".png", ".jpg"] and ext2.lower() in [".png", ".jpg"]:
        return True
    else:
        return False


# Assumes args has resize_up, resize_down, and resize_first
def get_resize_enum(args):
    if args.resize_up:
        return ims.Resize.RESIZE_UP
    elif args.resize_down:
        return ims.Resize.RESIZE_DOWN
    elif args.resize_first:
        return ims.Resize.RESIZE_FIRST
    else:
        return ims.Resize.RESIZE_NONE


class IsPlural:
    def __init__(self, pl):
        if hasattr(pl, '__len__') and (not isinstance(pl, str)):
            self.is_plural = len(pl) > 1
        elif isinstance(pl, numbers.Number):
            self.is_plural = abs(pl) > 1
        else:
            self.is_plural = False


class PluralizableString:
    def __init__(self, base_string: str, non_plural_end: str, plural_end: str, pluralizable: IsPlural):
        self.text: str = base_string + (plural_end if pluralizable.is_plural else non_plural_end)

    def __repr__(self):
        return self.text


# Checks that all file paths are correct and that no conflicting variables exist.
def validate_arguments(args):
    if args.dirs_in is not None and args.files_in is not None:
        raise EnvironmentError("Can only specify -dirs_in or -files_in")
    if args.dirs_in is not None and (args.to_img or args.to_imgs) and not args.ext_in:
        raise EnvironmentError("No extension specified for images. ")
    if args.dirs_in is not None:
        missing_dirs = fo.check_dirs(args.dirs_in, return_missing=True)
        if len(missing_dirs) != 0:
            pl = IsPlural(missing_dirs)
            dir_str = PluralizableString("director", "y", "ies", pl)
            do_str = PluralizableString("do", "es", "", pl)
            raise EnvironmentError("The %s specified at %s %s not exist." % (dir_str, ", ".join(missing_dirs), do_str))

    if args.files_in is not None:
        if len(args.files_in) < args.cols * args.rows:
            raise EnvironmentError(
                "Not enough photos given to generate an image or video of dimensions %d by %d (requires %d images or "
                "videos, %d given)" % (args.cols, args.rows, args.cols * args.rows, len(args.files_in)))

        first_ext = os.path.splitext(args.files_in[0])
        for path in args.files_in:
            if not os.path.exists(path):
                raise EnvironmentError("File %s not found." % path)
            if not check_ext(first_ext[1], os.path.splitext(path)[1]):
                raise EnvironmentError("Video and image files can not be included.")

        missing_files = fo.check_files(args.files_in, return_missing=True)
        if len(missing_files) != 0:
            file_str = PluralizableString("file", "", "s", IsPlural(missing_files))
            raise EnvironmentError("The %s specified at %s does not exist." % (file_str, ", ".join(missing_files)))

    if (args.resize_up and args.resize_down) or (args.resize_up and args.resize_first) \
            or (args.resize_first and args.resize_down):
        raise EnvironmentError("Choose one of --resize_up, --resize_down, --resize_first.")
    if args.resize_in is not None and args.resize_out is not None:
        raise EnvironmentError("Choose one of --resize_in or --resize_out.")
    if (args.to_vid and args.to_img) or (args.to_img and args.to_imgs):
        raise EnvironmentError("--to_img is not compatible with either --to_vid or --to_imgs.")



from enum import Enum


class Channel(Enum):
    RED = 1
    BLUE = 2
    GREEN = 3
    ALL = 4


# Input paths should be video files. Output should be full path.
# Will stack or videos in directory x by y, and resize each input image to width by height px.
def make_random_vid(output_path, width=64, height=48, num_frames=100, fps=24):
    import cv2
    fformat = 'mp4v'
    fformat = cv2.VideoWriter_fourcc(*fformat)

    vid = cv2.VideoWriter(filename=output_path, apiPreference=0, fourcc=fformat, fps=fps,
                          frameSize=(width, height))
    for i in range(num_frames):
        vid.write(make_static_img(width, height))
    vid.release()


def make_static_img(width, height, channels=[Channel.ALL]):
    from numpy.random import randint
    from numpy import zeros
    if Channel.ALL in channels:
        return randint(0, 255, (height, width, 3)).astype('uint8')
    arr1 = zeros((height, width, 3))
    if Channel.BLUE in channels:
        arr1[:, :, 0] = randint(0, 255, (height, width))
    if Channel.GREEN in channels:
        arr1[:, :, 1] = randint(0, 255, (height, width))
    if Channel.RED in channels:
        arr1[:, :, 2] = randint(0, 255, (height, width))
    return arr1.astype('uint8')


def save_static_img(output_path, width, height, channels=Channel.ALL):
    import cv2
    if isinstance(channels, list):
        cv2.imwrite(output_path, make_static_img(width, height, channels))
    else:
        cv2.imwrite(output_path, make_static_img(width, height, [channels]))



def get_correct_dimensions(args):
    if args.resize_in:
        return args.resize_in
    elif args.resize_out:
        image_width, image_height = args.resize_out
        if image_width % args.cols != 0:
            raise EnvironmentError("Output width must be a multiple of image stacking number.")
        if image_height % args.rows != 0:
            raise EnvironmentError("Output height must be a multiple of image stacking number.")
        image_width //= args.cols
        image_height //= args.rows
        return image_width, image_height

    if args.dirs_in is not None:
        return ims.get_dimensions_dirs(args.dirs_in, args.ext_in, ap.get_resize_enum(args))
    else:
        return ims.get_dimensions_files(args.files_in, args.ext_in, ap.get_resize_enum(args))


if __name__ == '__main__':
    import arg_parser as ap
    import videostacker as vs
    import imagestacker as ims
    import file_ops as fo

    import os

    args = ap.parse_arguments()

    os.makedirs(os.path.dirname(args.dir_out), exist_ok=True)

    if args.read_matching_file_names:
        image_width, image_height = None, None
        if args.resize_in:
            image_width, image_height = args.resize_in
        elif args.resize_out:
            image_width, image_height = args.resize_out
        ims.make_images_from_folders_match(args.dirs_in, args.dir_out, args.max_imgs,
                                           args.cols, args.rows, ap.get_resize_enum(args), image_width, image_height)
        exit()

    image_width, image_height = get_correct_dimensions(args)

    print("Output file will have dimensions: %d x %d px." % (image_width * args.cols, image_height * args.rows,))

    if args.to_vid:
        print(args)
        if args.dirs_in is not None and args.ext_in not in ["mp4"]:
            vs.make_video_from_images(args.dirs_in, args.ext_in, args.dir_out, args.name, args.ext_out, fps=args.fps,
                                      cols=args.cols, rows=args.rows, width=image_width, height=image_height)
        elif args.files_in is not None:
            vs.make_video_from_videos(args.files_in, args.dir_out, args.name, args.ext_out, cols=args.cols,
                                      rows=args.rows, width=image_width, height=image_height)
        else:
            files_in = fo.get_first_n_files(args.dirs_in, args.ext_in, args.cols * args.rows)
            if len(files_in) != args.cols * args.rows:
                raise ValueError("Insufficient files found in %s" % ", ".join(args.dirs_in))
            print("Automatically selected these video files to concatenate: %s" % (", ".join(files_in)))
            vs.make_video_from_videos(files_in, args.dir_out, args.name, args.ext_out,
                                      cols=args.cols, rows=args.rows, width=image_width, height=image_height)

    else:
        if args.files_in is not None:
            ims.make_image_from_images(args.files_in, args.dir_out, args.name, args.ext_out,
                                       cols=args.cols, rows=args.rows, width=image_width, height=image_height)
        else:
            ims.make_images_from_folders(args.dirs_in, args.ext_in, args.dir_out, args.name,
                                         ext_out=args.ext_out, max_imgs=args.max_imgs, rows=args.rows, cols=args.cols,
                                         width=image_width, height=image_height)



from typing import Union, List


def check_files(files: List[str], return_missing: bool = False) -> Union[List[str], bool]:
    """" If return_missing is true, returns an array of any missing files, otherwise, returns a boolean indicating
    that all of the files exist.

    Keyword arguments:
    files -- one or more files to check.
    return_missing -- return the missing files (default False).
    """

    missing = []
    import os
    if isinstance(files, str):
        if os.path.exists(files) and os.path.isfile(files):
            return [] if return_missing else True
        return [files] if return_missing else False
    else:
        for file in files:
            if not os.path.exists(file) or not os.path.isfile(file):
                if not return_missing:
                    return False
                else:
                    missing.append(file)
        if not return_missing:
            return True
        else:
            return missing


def check_dirs(directories: List[str], return_missing: bool = False) -> Union[List[str], bool]:
    """" If return_missing is true, returns an array of any missing directories, otherwise, returns a boolean indicating
    that all of the directories exist.

    Keyword arguments:
    directories -- one or more directories to check.
    return_missing -- return the missing directories (default False).
    """

    missing = []
    import os
    if isinstance(directories, str):
        if os.path.exists(directories) and os.path.isdir(directories):
            return [] if return_missing else True
        return [directories] if return_missing else False
    else:
        for directory in directories:
            if not os.path.exists(directory) or not os.path.isdir(directory):
                if not return_missing:
                    return False
                else:
                    missing.append(directory)
        if not return_missing:
            return True
        else:
            return missing


def get_files(directory: str, ext: Union[List[str], str]) -> List[str]:
    import glob
    """" Returns a list of file names in the given directory ending in the given extensions

    Keyword arguments:
    directories -- one or more directories to search.
    ext -- one or more extensions to match.
    """
    jpg_subset = [".JPG", ".jpeg", ".JPEG", ".jpg"]
    png_subset = [".PNG", ".png"]
    if ext is None:
        extensions = jpg_subset + png_subset
    elif isinstance(ext, str):
        if "jpg" in ext.lower():
            extensions = jpg_subset
        elif "png" in ext.lower():
            extensions = png_subset
        elif "mp4" in ext.lower():
            extensions = ['.mp4', '.MP4']
        else:
            extensions = [ext]
    else:
        extensions = ext

    extensions = [("." if extx[0] != "." else "") + extx for extx in extensions]
    directory = append_forward_slash_path(directory)
    frames = [frame for ext1 in extensions
              for frame in glob.glob(f"{directory}*{ext1}")]
    # sorts, removes duplicates.
    frames = sorted(list(set(frames)))
    return frames


# Gets the first n videos from the directories specified, as evenly as possibly.
def get_first_n_files(directories: Union[List[str], str], ext: Union[List[str], str], num: int) -> List[str]:
    """ Returns the first n files in the directories that match the given extensions, as evenly as possible.
    In the event that less than num files exist, will return all matches found.

    Keyword arguments:
    directories -- one or more directories to search.
    ext -- one or more extensions to match.
    num -- number of files that should be matched and returned.
    """

    if not isinstance(directories, list):
        directories = [directories]
    dirs = []
    for directory in directories:
        dirs.append(get_files(append_forward_slash_path(directory), ext))
    dir_exhausted = [False] * len(dirs)

    curr_dir = 0
    curr_index = 0
    output = []
    while len(output) < num:
        # Directory has no images left
        if curr_index >= len(dirs[curr_dir]):
            # Given directory is "exhausted": no images left
            dir_exhausted[curr_dir] = True
            # All directories are exhausted.
            if all(dir_exhausted):
                return output
        else:
            # Otherwise, append next image in directory.
            output.append(dirs[curr_dir][curr_index])
        # Move to next image in directory.
        curr_dir += 1
        if curr_dir % len(dirs) == 0:
            curr_index += 1
            curr_dir = 0
    return output


def append_forward_slash_path(paths: Union[List[str], str]) -> Union[List[str], str, None]:
    """ Returns the input string(s), in the same format as they were passed in, with a minimum of one forward slash
    at the end, given that no forward slash exists at the end.

    Keyword arguments:
    paths -- one or more paths to add a forward slash to.
    """
    if paths is None:
        return None
    if isinstance(paths, str):
        if paths[-1] != "/":
            print("Appended a / to %s" % paths)
            paths += "/"
        return paths
    else:
        output = [path + ("/" if path[-1] != "/" else "") for path in paths]
        return output


def clear_files(folder: str, *argv) -> None:
    """ Clears the given folder of one or more extensions.

    Keyword arguments:
    folder -- folder to remove extensions from.
    *argv -- a list of one or more extensions.
    """
    from os import remove
    from os import path
    from glob import glob
    folder = append_forward_slash_path(folder)
    for ext in argv:
        ext = ("." if ext[0] != "." else "") + ext
        for f in glob(path.join(folder) + "*" + ext):
            remove(f)


def form_file_name(dir_out: str, file_name: str, ext: str) -> str:
    """ Removes excess extensions in the file_name and returns a fully formed file name, cleaned of excess extensions.

    Keyword arguments:
    dir_out -- path to a directory.
    file_name -- a file name with zero or more extensions.
    ext -- the extension the file name should end in.
    """

    # Needs to check that file_name doesn't contain an extension.
    from os import path
    split_name = path.splitext(file_name)
    ext = ("." if ext[0] != "." else "") + ext
    if len(split_name) > 1:
        file_name = ".".join(split_name[:-1])
    return path.join(dir_out, file_name + ext)



import cv2
from enum import Enum
from typing import Union, List
import os

from . import file_ops as fo


class Resize(Enum):
    RESIZE_UP = 3
    RESIZE_DOWN = 4
    RESIZE_FIRST = 5
    RESIZE_NONE = 0


class ImageDataStore:
    def __init__(self, images, file_name, ext):
        self.images = images
        self.file_name = file_name
        self.ext = ext


class ImageGenerator:
    # TODO: allow multiple file extension inputs.
    # Pass in paths to directories containing images, and extension of desired image inputs. Will return num images each iteration.
    def __init__(self, directories, ext='jpg', num=1):
        self.directories = directories
        self.ext = ext
        self.num = num
        self.files = {}
        self.min_files = 0
        self.curr_index = 0
        self.active_dir = 0
        self.index = {}
        # Checks that all directories are valid.
        if fo.check_dirs(self.directories):
            self.load_dirs()
        else:
            raise EnvironmentError("One or more directories do not exist.")

    # Sets the number of files that this iterator will output:
    # If min_files <= 0, nothing will be returns. If min_files >= self.min_files, self.min_files remains unchanged.
    # Otherwise, self.min_files = min_files.
    def set_min_files(self, min_files: Union[None, int]):
        if min_files is not None:
            if min_files < 0:
                self.min_files = 0
            elif min_files <= self.min_files:
                self.min_files = min_files

    # Loads all image paths into memory.
    def load_dirs(self):
        if isinstance(self.directories, str):
            self.files[0] = fo.get_files(self.directories, self.ext)
            self.index[0] = 0
            self.min_files = len(self.files[0]) - len(self.files[0]) % self.num
            if self.min_files == 0:
                raise EnvironmentError("No images matching %s found in directories %s."
                                       % (", ".join(self.ext), ", ".join(self.directories)))

        else:
            counter = 0
            for directory in self.directories:
                files = fo.get_files(self.directories[counter], self.ext)
                if self.min_files == 0:
                    self.min_files = len(files)
                elif self.min_files > len(files):
                    self.min_files = len(files)
                if self.min_files == 0:
                    print("No files found in " + directory)
                else:
                    self.files[counter] = files
                    self.index[counter] = 0
                    counter += 1

            if self.min_files == 0:
                raise EnvironmentError("No images matching %s found in directories %s."
                                       % (", ".join(self.ext), ", ".join(self.directories)))
            self.min_files = len(self.files[0]) - len(self.files[0]) % self.num

    def __iter__(self):
        return self

    # Returns num images in array.
    def __next__(self):
        if self.curr_index < self.min_files * len(self.files.keys()):
            output = []
            for i in range(self.num):
                output.append(cv2.imread(self.files[self.active_dir][self.index[self.active_dir]]))
                self.index[self.active_dir] += 1
                self.curr_index += 1
                self.active_dir += 1
                self.active_dir %= len(self.files.keys())

            return ImageDataStore(output, None, None)
        else:
            raise StopIteration()


class ImageGeneratorMatchToName:
    # TODO: allow multiple file extension inputs.
    # Pass in paths to directories containing images.
    def __init__(self, directories):
        self.directories = directories
        self.files = {}
        self.num_imgs = 0
        self.curr_index = 0
        self.possible_file_names_arr = []
        # Checks that all directories are valid.
        if fo.check_dirs(self.directories):
            self.load_dirs()
        else:
            raise EnvironmentError("One or more directories do not exist.")

    def set_min_files(self, min_files: Union[None, int]):
        if min_files is not None:
            if min_files < 0:
                self.num_imgs = 0
            elif min_files <= self.num_imgs:
                self.num_imgs = min_files

    # Loads all image paths into memory.
    def load_dirs(self):
        import os
        possible_file_names = {}
        if isinstance(self.directories, str):
            self.files[0] = {
                os.path.basename(f_name): f_name
                for f_name in fo.get_files(self.directories, ["jpg", "png"])
            }
            possible_file_names = {key: True for key in self.files}
        else:
            counter = 0
            for directory in self.directories:
                files = fo.get_files(self.directories[counter], ["jpg", "png"])
                if files:
                    self.files[counter] = {os.path.basename(f_name): f_name
                                           for f_name in files}
                    possible_file_names = {key: (key in possible_file_names)
                                           for key in self.files[counter]}
                    counter += 1
                else:
                    raise ValueError(f"No files found in {directory}")
        self.possible_file_names_arr = [key for key, val in possible_file_names.items() if val]
        if self.possible_file_names_arr:
            self.num_imgs = len(self.possible_file_names_arr)
        else:
            raise ValueError("No file names in common.")

    def __iter__(self):
        return self

    # Returns num images in array and filename.
    def __next__(self):
        import os
        if self.curr_index < self.num_imgs:
            output = []
            file_name = self.possible_file_names_arr[self.curr_index]

            for key in self.files.keys():
                output.append(cv2.imread(self.files[key][file_name]))
            self.curr_index += 1

            return ImageDataStore(output, os.path.splitext(file_name)[0], os.path.splitext(file_name)[-1])
        else:
            raise StopIteration()


# Expects an array of images a tuple (x,y) that represents how the images will be stacked, and a mode representing
# how the array will be stacked:
# eg. images = [img]*6, dimensions = (2,3), mode='rd':
# 2 images per row, 3 rows, ordered from left to right, up to down
def stack_images(images, dimensions, mode='rd'):
    import numpy as np
    x = dimensions[0]
    y = dimensions[1]
    images_stacked = [None] * y
    for i in range(y):
        images_stacked[i] = [None] * x
    if mode[0] in ('l', 'r'):
        for i in range(x * y):
            images_stacked[i // x if mode[1] == 'd' else y - i // x - 1][i % x if mode[0] == 'r' else x - i % x - 1] = \
                images[i]
    elif mode[0] in ('u', 'd'):
        for i in range(x * y):
            images_stacked[i % y if mode[0] == 'd' else y - i % y - 1][i // y if mode[0] == 'r' else x - i // y - 1] = \
                images[i]
    return np.concatenate(tuple([np.concatenate(tuple(row), axis=1) for row in images_stacked]), axis=0)


# Resizes all of the images in the input to the specified dimensions.
def resize_images(images, dimensions: (int, int)):
    for i in range(len(images)):
        images[i] = cv2.resize(images[i], dimensions)
    return images


def make_image_from_images(files_in: Union[List[str], str], dir_out="./", file_name="output", ext_out="jpg",
                           cols: int = 1, rows: int = 1, width: int = 640, height: int = 480, mode: str = 'rd'):
    file_name = fo.form_file_name(dir_out, file_name, ext_out)
    os.makedirs(os.path.dirname(file_name), exist_ok=True)
    if isinstance(files_in, str):
        cv2.imwrite(file_name, cv2.imread(files_in))
    elif len(files_in) == 1:
        cv2.imwrite(file_name, cv2.imread(files_in[0]))
    else:
        images = []
        for file in files_in:
            images.append(cv2.imread(file))
        cv2.imwrite(file_name, stack_images(resize_images(images, (width, height)), (cols, rows), mode=mode))


def make_images_from_folders_match(dirs_in, dir_out, max_imgs=None, cols=1, rows=1,
                                   resize_opt: Resize = Resize.RESIZE_FIRST, width=None, height=None, mode='rd'):
    image_iter = ImageGeneratorMatchToName(dirs_in)
    counter: int = 0
    fn = return_first

    if resize_opt == Resize.RESIZE_UP:
        fn = return_max
    elif resize_opt == Resize.RESIZE_DOWN:
        fn = return_min

    image_iter.set_min_files(max_imgs)
    os.makedirs(dir_out, exist_ok=True)
    for image_data in image_iter:
        if max_imgs is None or counter < max_imgs:
            images = image_data.images
            file_name = fo.form_file_name(dir_out, image_data.file_name, image_data.ext)
            if width is None or height is None:
                dimensions = [(image.shape[1], image.shape[0]) for image in images]
                temp_width, temp_height = fn(dimensions)
                cv2.imwrite(file_name, stack_images(resize_images(images, (temp_width, temp_height)),
                                                    (cols, rows), mode=mode))
            else:
                cv2.imwrite(file_name, stack_images(resize_images(images, (width, height)), (cols, rows), mode=mode))
            counter += 1
        else:
            break


# Draws images with the given extension equally from each folder,
#  resizes each individual image to width x height, concatenates them according to the mode, and saves them to the
def make_images_from_folders(dirs_in: Union[List[str], str], ext_in: Union[List[str], str] = 'jpg',
                             dir_out: str = "./", file_name: str = "output", ext_out='jpg', max_imgs: int = None,
                             cols: int = 1, rows: int = 1, width: int = 640, height: int = 480, mode='rd'):
    image_iter = ImageGenerator(dirs_in, ext=ext_in, num=cols * rows)
    os.makedirs(dir_out, exist_ok=True)
    image_iter.set_min_files(max_imgs)
    num_zeros = len(str(image_iter.min_files // image_iter.num -1))
    for counter, images_data in enumerate(image_iter):
        images = images_data.images
        temp_file_name = fo.form_file_name(dir_out, file_name + str(counter).zfill(num_zeros), ext_out)
        cv2.imwrite(temp_file_name, stack_images(resize_images(images, (width, height)), (cols, rows), mode=mode))


# Returns the appropriate dimensions, given a set of
def get_dimensions_dirs(dirs_in: Union[List[str], str], ext: str, resize: Resize):
    if resize == Resize.RESIZE_FIRST:
        return get_first_dimensions_dirs(dirs_in, ext)
    if resize == Resize.RESIZE_UP:
        return get_max_dimensions_dirs(dirs_in, ext)
    if resize == Resize.RESIZE_DOWN:
        return get_min_dimensions_dirs(dirs_in, ext)
    else:
        return get_first_dimensions_dirs(dirs_in, ext)


def get_dimensions_files(files_in: Union[List[str], str], ext: str, resize: Resize):
    if resize == Resize.RESIZE_FIRST:
        return get_first_dimensions_files(files_in, ext)
    if resize == Resize.RESIZE_UP:
        return get_max_dimensions_files(files_in, ext)
    if resize == Resize.RESIZE_DOWN:
        return get_min_dimensions_files(files_in, ext)
    else:
        return get_first_dimensions_files(files_in, ext)


# Will get the dimensions of the first file in dirs_in with ext_in.
def get_first_dimensions_dirs(dirs_in: Union[List[str], str], ext_in: Union[List[str], str]):
    if len(dirs_in) == 0:
        raise Exception("Insufficient directories.")

    if isinstance(dirs_in, str):
        return get_first_dimensions_files(fo.get_files(dirs_in, ext_in), ext_in)
    else:
        for dir_in in dirs_in:
            try:
                return get_first_dimensions_files(fo.get_files(dir_in, ext_in), ext_in)
            except ValueError:
                continue
        if isinstance(ext_in, str):
            ext_in = [ext_in]
        raise ValueError("No files with given extension %s found in any directory." % (", ".join(ext_in),))


# Returns the range of image sizes.
def get_min_dimensions_dirs(dirs_in: Union[List[str], str], ext_in: Union[List[str], str]):
    if len(dirs_in) == 0:
        raise Exception("Insufficient directories.")

    if isinstance(dirs_in, str):
        return get_min_dimensions_files(fo.get_files(dirs_in, ext_in), ext_in)
    else:
        dims = []
        for dir_in in dirs_in:
            try:
                dims.append(get_min_dimensions_files(fo.get_files(dir_in, ext_in), ext_in))
            except ValueError:
                continue
        try:
            return return_min(dims)
        except ValueError:
            raise ValueError("No files with given extension %s found in any directory." % (ext_in,))


# Returns the range of image sizes.
def get_max_dimensions_dirs(dirs_in: Union[List[str], str], ext_in: Union[List[str], str]):
    if len(dirs_in) == 0:
        raise Exception("Insufficient directories.")

    if isinstance(dirs_in, str):
        return get_max_dimensions_files(fo.get_files(dirs_in, ext_in), ext_in)
    else:
        dims = []
        for dir_in in dirs_in:
            try:
                dims.append(get_max_dimensions_files(fo.get_files(dir_in, ext_in), ext_in))
            except ValueError:
                continue
        try:
            return return_max(dims)
        except ValueError:
            raise ValueError("No files with given extension %s found in any directory." % (ext_in,))


# Will get the dimensions of the first file in files_in
def get_first_dimensions_files(files_in: Union[List[str], str], ext_in: Union[List[str], str]):
    if isinstance(files_in, str):
        files_in = [files_in]

    if len(files_in) == 0:
        raise ValueError("Insufficient files.")

    if "mp4" not in ext_in:
        return return_first(get_img_dimensions(files_in))
    else:
        return return_first(get_video_dimensions(files_in))


# Gets the minimum and maximum dimensions of the files in the list.
def get_min_dimensions_files(files_in: Union[List[str], str], ext_in: Union[List[str], str]):
    if isinstance(files_in, str):
        files_in = [files_in]

    if len(files_in) == 0:
        raise ValueError("Insufficient files.")

    if "mp4" not in ext_in:
        return return_min(get_img_dimensions(files_in))
    else:
        return return_min(get_video_dimensions(files_in))


# Gets the minimum and maximum dimensions of the files in the list with the given extension(s).
def get_max_dimensions_files(files_in: Union[List[str], str], ext_in: Union[List[str], str]):
    if isinstance(files_in, str):
        files_in = [files_in]

    if len(files_in) == 0:
        raise ValueError("Insufficient files.")

    if "mp4" not in ext_in:
        return return_max(get_img_dimensions(files_in))
    else:
        return return_max(get_video_dimensions(files_in))


# Given a list of file names, returns the dimensions of the images corresponding to the file names.
def get_img_dimensions(imgs_in: Union[List[str], str]):
    dims_list = []
    for img in imgs_in:
        file = cv2.imread(img)
        dims_list.append((file.shape[1], file.shape[0]))

    return dims_list


# Given a list of file names, returns the dimensions of the videos corresponding to the file names.
def get_video_dimensions(videos_in: Union[List[str], str]):
    dims_list = []
    for vid in videos_in:
        file = cv2.VideoCapture(vid)
        dims_list.append((int(file.get(cv2.CAP_PROP_FRAME_WIDTH)), int(file.get(cv2.CAP_PROP_FRAME_HEIGHT))))
        file.release()

    return dims_list


# Returns the first element in a given list.
def return_first(list_of_dims):
    return list_of_dims[0] if len(list_of_dims) > 0 else None


# Returns the dimensions that produce the maximum area. In the event of a tie, will return the first match.
def return_max(list_of_dims):
    return max(list_of_dims, key=lambda dim: dim[0] * dim[1])


# Returns the dimensions that produce the minimum area. In the event of a tie, will return the first match.
def return_min(list_of_dims):
    return min(list_of_dims, key=lambda dim: dim[0] * dim[1])



from typing import Union, List

import cv2
import os

from . import file_ops as fo
from . import imagestacker as ims


class VideoIterator:
    def __init__(self, paths_to_videos, num=1):
        """ Initializes a video iterator that will return num frames per iteration from each video in paths_to_videos.

        :param paths_to_videos:
        :param num:
        """
        if not fo.check_files(paths_to_videos):
            raise EnvironmentError("One or more videos not found.")

        self.num = num
        self.videos = {}
        self.paths = paths_to_videos
        self.videos_completed = {}
        self.last_frame = {}
        self.active_vid = 0
        self.fps = 0
        self.width = None
        self.height = None
        self.num_frames = 0
        self.load_videos()

    # Will find the smallest video dimensions Assumes that all videos have been initialized.
    def find_min_dims(self):
        if self.height is None or self.width is None:
            dims = []
            for video in self.videos.values():
                dims.append((int(video.get(cv2.CAP_PROP_FRAME_WIDTH)), int(video.get(cv2.CAP_PROP_FRAME_HEIGHT))))
            self.width, self.height = ims.return_min(dims)

    # Loads the videos into memory, initializes variables.
    def load_videos(self):
        if isinstance(self.paths, str):
            self.last_frame[0] = None
            self.videos[0] = cv2.VideoCapture(self.paths)
            self.width = int(self.videos[0].get(cv2.CAP_PROP_FRAME_WIDTH))
            self.height = int(self.videos[0].get(cv2.CAP_PROP_FRAME_HEIGHT))
            self.fps = self.videos[0].get(cv2.CAP_PROP_FPS)
            self.num_frames = int(self.videos[0].get(cv2.CAP_PROP_FRAME_COUNT))
            self.videos_completed[0] = False
        else:
            counter = 0
            for video in self.paths:
                self.last_frame[counter] = None
                self.videos[counter] = cv2.VideoCapture(video)
                self.videos_completed[counter] = False
                if self.fps != 0:
                    if int(self.fps) != int(self.videos[counter].get(cv2.CAP_PROP_FPS)):
                        raise EnvironmentError("Video FPS does not match.")
                else:
                    self.fps = self.videos[counter].get(cv2.CAP_PROP_FPS)
                self.num_frames = max(self.num_frames, int(self.videos[0].get(cv2.CAP_PROP_FRAME_COUNT)))
                counter += 1
            self.find_min_dims()

    def __iter__(self):
        return self

    # Returns num frames from all videos. If one video has reached the end, will keep last frame.
    def __next__(self):
        if not all(val is True for val in self.videos_completed.values()):
            output = []
            for i in range(self.num):
                if not self.videos_completed[self.active_vid]:
                    success, image = self.videos[self.active_vid].read()
                    if success:
                        self.last_frame[self.active_vid] = image
                    else:
                        self.videos_completed[self.active_vid] = True
                output.append(self.last_frame[self.active_vid])
                self.active_vid += 1
                self.active_vid %= len(self.videos.keys())
            return output
        else:
            raise StopIteration()


# ORDER OF ARGUMENTS:
# input_paths, [input file types], output_path, [output file types], [num_imgs]
#  cols, rows, [dimension altering], width, height, mode

# TODO: add support for more video formats
def make_video_from_images(dirs_in, ext_in="jpg", dir_out="./", file_name="output", ext_out="mp4",
                           video_format='mp4v', fps=24, cols=1, rows=1, width=None, height=None):
    """Creates, and saves a video with the file_name, encoded using video_format containing each video in files_in,
    containing each image in dirs_in with the appropriate extension, in order of appearance, with each individual image
    resized to width, height, stacked col x row.

    eg. dirs_in a b c d
    cols 2
    rows 2
    frame = a[i] b[i]
            c[i] d[i]
    where x[i] refers to the ith file in that directory

    :param dirs_in:         List of files to read and place into the video.
    :param ext_in:          choose files in the directory with the given extension(s).
    :param dir_out:         directory to output the file to. If it does not exist, it will be created automatically.
    :param file_name:       name of output video
    :param ext_out:         output extension for the video (default mp4)
    :param video_format:    format to encode the video in (default mp4v)
    :param fps:             desired fps of output video.
    :param cols:            number of images placed side by side (default 1)
    :param rows:            number of images placed vertically (default 1)
    :param width:           width of each sub component in px (resulting video will be width * col px high).
    :param height:          height of each sub component in px (resulting video will be height * row px high).
    :return:                nothing
    """

    supported_extensions = ["mp4"]
    if ext_out not in supported_extensions:
        raise ValueError("Extension %s is not currently supported." % (ext_out,))

    video_format = cv2.VideoWriter_fourcc(*video_format)
    # apiPreference may be required depending on cv2 version.
    image_iter = ims.ImageGenerator(dirs_in, ext_in, cols * rows)
    vid = cv2.VideoWriter(filename=fo.form_file_name(dir_out, file_name, ext_out),
                          apiPreference=0, fourcc=video_format, fps=fps, frameSize=(width * cols, height * rows))
    for images_data in image_iter:
        vid.write(ims.stack_images(ims.resize_images(images_data.images, (width, height)), (cols, rows)))
    vid.release()


def make_video_from_array(files_in, dir_out="./", file_name="output", ext_out="mp4",
                          video_format='mp4v', fps=24, width=None, height=None):
    """ Creates, and saves a video with the file_name, encoded using video_format containing each video in files_in,
    containing each image in files_in in order of appearance, with each individual image resized to width, height.

    :param files_in:        List of files to read and place into the video.
    :param dir_out:         directory to output the file to. If it does not exist, it will be created automatically.
    :param file_name:       name of output video
    :param ext_out:         output extension for the video (default mp4)
    :param video_format:    format to encode the video in (default mp4v)
    :param fps:             desired fps of output video.
    :param width:           width of each sub component in px (resulting video will be width * col px high).
    :param height:          height of each sub component in px (resulting video will be height * row px high).
    :return:                nothing
    """

    supported_extensions = ["mp4"]
    if ext_out not in supported_extensions:
        raise ValueError("Extension %s is not currently supported." % (ext_out,))

    video_format = cv2.VideoWriter_fourcc(*video_format)
    # apiPreference may be required depending on cv2 version.
    if isinstance(files_in, str):
        files_in = [files_in]
    exts = [x for x in set([os.path.splitext(file_name)[1:] for file_name in files_in])]
    if width is None or height is None:
        width, height = ims.get_first_dimensions_files(files_in, exts)
    vid = cv2.VideoWriter(filename=fo.form_file_name(dir_out, file_name, ext_out),
                          apiPreference=0, fourcc=video_format, fps=fps, frameSize=(width, height))
    for image in files_in:
        im = cv2.imread(image)
        vid.write(ims.resize_images([im], (width, height))[0])
    vid.release()


# Input paths should be video files. Output should be full path.
# Will stack or videos in directory x by y, and resize each input image to width by height px.
def make_video_from_videos(files_in: Union[List[str], str], dir_out: str = "./", file_name: str = "output",
                           ext_out: str = "mp4", video_format: str = 'mp4v',
                           cols: int = 1, rows: int = 1, width: int = None, height: int = None) -> None:
    """ Creates, and saves a video with the file_name, encoded using video_format containing each video in files_in,
    stacked col x row, in order of appearance, with each individual video frame resized to width, height.

    :param files_in:        List of files to read and place into the video.
    :param dir_out:         directory to output the file to. If it does not exist, it will be created automatically
    :param file_name:       Name of output video
    :param ext_out:         output extension for the video (default mp4)
    :param video_format:    format to encode the video in (default mp4v)
    :param cols:            number of images placed side by side (default 1)
    :param rows:            number of images placed vertically (default 1)
    :param width:           width of each sub component in px (resulting video will be width * col px high).
    :param height:          height of each sub component in px (resulting video will be height * row px high).
    :return:                nothing
    """

    supported_extensions = ["mp4"]
    if ext_out not in supported_extensions:
        raise ValueError("Extension %s is not currently supported." % (ext_out,))

    video_format = cv2.VideoWriter_fourcc(*video_format)
    video_iter = VideoIterator(files_in, cols * rows)

    if width is None:
        width = video_iter.width
    if height is None:
        height = video_iter.height

    vid = cv2.VideoWriter(filename=fo.form_file_name(dir_out, file_name, ext_out), apiPreference=0,
                          fourcc=video_format, fps=video_iter.fps, frameSize=(width * cols, height * rows))
    for images in video_iter:
        vid.write(ims.stack_images(ims.resize_images(images, (width, height)), (cols, rows)))
    vid.release()


# Splits video into a given number of frames, or all frames if frame_count = -1, and saves frames to output_dir,
# with sequential filenames (eg. 0000.png, 0001.png ... 9999.png, 0000.png is frame #1)
def split_video(file_in: str, dir_out: str, file_name: str = "", ext_out: str = "png", frame_count: int = -1,
                start_frame: int = 0, end_frame: int = -1):
    """ Takes each individual frame from the video file_in, and outputs it as an image with the file_name followed by a
    padded counter corresponding to that image's position in the video (eg. 0001) to dir_out. ext_out,
    starting at start_frame and going to end_frame. Outputs frame_count stills if end_frame is not specified.

    :param file_in:         List of files to read and place into the video.
    :param dir_out:         directory to output the file to. If it does not exist, it will be created automatically.
    :param file_name:       first part of output file names.
    :param ext_out:         output extension for the stills
    :param frame_count:     number of frames to save, starting at start frame. Overrides end_frame. (default -1, or all)
    :param start_frame:     first frame of video to save (default 0, beginning of video)
    :param end_frame:       frame of video to save to (not including) (default -1, end of video)
    :return:                nothing
    """

    import os
    ext_out = ("" if ext_out[0] == "." else ".") + ext_out
    os.makedirs(dir_out, exist_ok=True)
    vid_iterator = VideoIterator(file_in)
    num_frames: int = vid_iterator.num_frames
    if end_frame != -1 and frame_count == -1:
        frame_count = min(end_frame, num_frames) - start_frame
    else:
        frame_count: int = min(frame_count, num_frames) if frame_count != -1 else num_frames
    if frame_count < 1:
        raise ValueError("Values passed in result in no or negative frames of output.")
    num_zeros = len(str(frame_count - 1))
    for frame, counter in zip(vid_iterator, range(0, num_frames)):
        if frame_count > counter - start_frame >= 0:
            temp_name = os.path.join(dir_out,
                                     f"{file_name}{str(counter - start_frame).zfill(num_zeros)}{ext_out}")
            print("saving to %s" % temp_name)
            cv2.imwrite(temp_name, frame[0])
        elif counter > start_frame + frame_count:
            return
